import logging
import threading
from functools import _make_key, wraps

from ocyan.core.fender import config

logger = logging.getLogger(__name__)

DISABLE_METHOD_REQUEST_CACHE = config.get("core", "disable_method_request_cache", False)
DISABLE_TEMPLATETAG_REQUEST_CACHE = config.get(
    "core", "disable_templatetag_request_cache", False
)

__all__ = ["RequestCache", "RequestLibrary"]


class RequestCache(threading.local):
    CACHE_STORAGE_REQUEST_KEY = "_request_cache"

    class CacheForKey(object):
        def __init__(self, cache, cache_key):
            self.cache = cache
            self.cache_key = cache_key

        def __bool__(self):
            return self.cache_key in self.cache

        def get(self):
            return self.cache.get(self.cache_key)

        def set(self, val):
            self.cache.set(self.cache_key, val)

        def clear(self):
            self.cache.delete(self.cache_key)

    @classmethod
    def cache_for_request(cls, request):
        if not hasattr(request, cls.CACHE_STORAGE_REQUEST_KEY):
            setattr(request, cls.CACHE_STORAGE_REQUEST_KEY, cls())

        return getattr(request, cls.CACHE_STORAGE_REQUEST_KEY)

    @classmethod
    def cache_for_key(cls, request, *args, **kwargs):
        """
        create a request cache object for a single key

        >>> import argparse
        >>> request = argparse.Namespace()
        >>> c = RequestCache.cache_for_key(request, "harrie")
        >>> bool(c)
        False
        >>> c.set(1)
        >>> bool(c)
        True
        >>> c.get()
        1
        >>> c.set("klaas")
        >>> bool(c)
        True
        >>> c.get()
        'klaas'
        >>> c.clear()
        >>> bool(c)
        False
        >>> c.get()
        >>> c.set("adje")
        >>> d = RequestCache.cache_for_key(None, "harrie")
        >>> bool(d)
        False
        >>> d.set(1)
        >>> bool(d)
        True
        >>> d.get()
        1
        >>> f = RequestCache.cache_for_key(request, "harrie")
        >>> bool(f)
        True
        >>> f.get()
        'adje'
        """
        cache_key = cls.make_key(*args, **kwargs)
        if request is None:
            return cls.CacheForKey(cls(), cache_key)

        return cls.CacheForKey(cls.cache_for_request(request), cache_key)

    def clear(self):
        self.__dict__.clear()

    def __contains__(self, value):
        return value in self.__dict__

    def set(self, name, value):
        self.__dict__[name] = value

    def get(self, name):
        return self.__dict__.get(name)

    def delete(self, name):
        self.__dict__.pop(name, None)

    @classmethod
    def make_key(cls, *args, **kwargs):
        return _make_key(args, kwargs, False)

    @classmethod
    def once_per_request(cls, method):
        if DISABLE_METHOD_REQUEST_CACHE:
            return method

        @wraps(method)
        def _wrapper(self, *args, **kwargs):
            if self.request is not None:
                cache = cls.cache_for_request(self.request)
                method_name = getattr(method, "_decorated_function", method).__name__

                if hasattr(self, "make_key"):
                    key = self.make_key(method_name, *args, **kwargs)
                else:
                    key = cache.make_key(method_name, *args, **kwargs)

                if key not in cache:
                    value = method(self, *args, **kwargs)
                    cache.set(key, value)
                    return value

                return cache.get(key)

            return method(self, *args, **kwargs)

        return _wrapper


class ContextRequestTemplateTag(object):
    """
    Templatetag that is forced to take context to read the request from
    """

    def __init__(self, tag_decorator, *args, takes_context=None, name=None):
        self.func_takes_context = takes_context
        self.tag_decorator = tag_decorator(*args, takes_context=True, name=name)

    def __call__(self, func):
        def contextrequesttemplatetag_wrapper(context, *args, **kwargs):
            request = context.get("request")
            if request is None:
                logger.warning(
                    "Useless application of RequestLibrary, no request in context"
                )
                if self.func_takes_context:
                    return func(context, *args, **kwargs)
                else:
                    return func(*args, **kwargs)

            # we are good, compute cache key and try cached value
            cache = RequestCache.cache_for_request(request)
            func_name = getattr(func, "_decorated_function", func).__name__
            key = cache.make_key(func_name, *args, **kwargs)

            if key not in cache:
                if self.func_takes_context:
                    value = func(context, *args, **kwargs)
                else:
                    value = func(*args, **kwargs)
                cache.set(key, value)
                return value  # immediately return the value

            return cache.get(key)

        return self.tag_decorator(contextrequesttemplatetag_wrapper)


class RequestTemplateTag(object):
    """
    Templatetag that take request as the first argument
    """

    def __init__(self, tag_decorator, *args, takes_context=None, name=None):
        assert (
            not takes_context
        ), "RequestTemplateTag can not take request and context at the same time"
        self.tag_decorator = tag_decorator(*args, takes_context=False, name=name)

    def __call__(self, func):
        def requesttemplatetag_wrapper(request, *args, **kwargs):
            cache = RequestCache.cache_for_request(request)
            func_name = getattr(func, "_decorated_function", func).__name__
            key = cache.make_key(func_name, *args, **kwargs)

            if key not in cache:
                value = func(request, *args, **kwargs)
                cache.set(key, value)
                return value

            return cache.get(key)

        return self.tag_decorator(requesttemplatetag_wrapper)


class RequestLibrary(object):
    def __init__(self, library):
        self.library = library

    def simple_tag(self, func=None, takes_context=None, name=None, takes_request=False):
        if DISABLE_TEMPLATETAG_REQUEST_CACHE:
            return func

        if takes_request:
            tag = RequestTemplateTag(
                self.library.simple_tag, takes_context=takes_context, name=name
            )
        else:
            tag = ContextRequestTemplateTag(
                self.library.simple_tag, takes_context=takes_context, name=name
            )

        if func is not None:
            return tag(func)
        return tag

    def inclusion_tag(
        self, filename, func=None, takes_context=None, name=None, takes_request=False
    ):
        if DISABLE_TEMPLATETAG_REQUEST_CACHE:
            return func

        if takes_request:
            tag = RequestTemplateTag(
                self.library.inclusion_tag,
                filename,
                takes_context=takes_context,
                name=name,
            )
        else:
            tag = ContextRequestTemplateTag(
                self.library.inclusion_tag, takes_context=takes_context, name=name
            )

        if func is not None:
            return tag(func)

        return tag



from .appextension import *  # noqa: F403
from .cache import *  # noqa: F403
from .csp import *  # noqa: F403
from .currency import *  # noqa: F403
from .dirtyhacks import *  # noqa: F403
from .fixture import *  # noqa: F403
from .itertools import *  # noqa: F403
from .rudiments import *  # noqa: F403
from .theme import *  # noqa: F403
from .typeconversion import *  # noqa: F403

# this is the old name :) a lot of code uses it
from .typeconversion import str_to_identifier as str_to_idenifier  # noqa: F401



__all__ = ["chunked", "unique_everseen", "unique_list", "find_first", "merge_dicts"]

from collections.abc import Iterable, Mapping
from itertools import chain, filterfalse, islice


def chunked(qs, size, startindex=0):
    """
    Divide an interable into chunks of ``size``

    >>> list(chunked("hahahaha", 2))
    ['ha', 'ha', 'ha', 'ha']
    >>> list(chunked([1,2,3,4,5,6,7], 3))
    [[1, 2, 3], [4, 5, 6], [7]]
    """
    while True:
        chunk = qs[startindex : startindex + size]
        chunklen = len(chunk)
        if chunklen:
            yield chunk
        if chunklen < size:
            break
        startindex += size


# from https://docs.python.org/3/library/itertools.html#itertools-recipes


def ichunked(qs, size):
    """
    Does the same as chunked, but works with arbitrary generators

    >>> list(map(lambda x: "".join(x), ichunked("hahahaha", 2)))
    ['ha', 'ha', 'ha', 'ha']
    >>> list(map(list, ichunked([1,2,3,4,5,6,7], 3)))
    [[1, 2, 3], [4, 5, 6], [7]]
    """
    iterator = iter(qs)
    for first in iterator:
        yield chain([first], islice(iterator, size - 1))


def unique_everseen(iterable, key=None):
    """List unique elements, preserving order. Remember all elements ever seen.

    >>> list(unique_everseen('AAAABBBCCDAABBB'))
    ['A', 'B', 'C', 'D']
    >>> list(unique_everseen('ABBCcAD', str.lower))
    ['A', 'B', 'C', 'D']
    """
    seen = set()
    seen_add = seen.add
    if key is None:
        for element in filterfalse(seen.__contains__, iterable):
            seen_add(element)
            yield element
    else:
        for element in iterable:
            k = key(element)
            if k not in seen:
                seen_add(k)
                yield element


def unique_list(iterable, key=None):
    """Return a list without duplicates, preserving order.

    >>> unique_list('AAAABBBCCDAABBB')
    ['A', 'B', 'C', 'D']
    >>> unique_list('ABBCcAD', str.lower)
    ['A', 'B', 'C', 'D']
    >>> unique_list([1, 3, 5, 3, 1, 7, 20])
    [1, 3, 5, 7, 20]
    """
    unique_iterable = unique_everseen(iterable, key=key)
    return list(unique_iterable)


def find_first(function, iterable):
    """
    Find first item in iterable that fulfills function

    >>> is_henk_1 = lambda x: x.get('henk') == 1
    >>> data = [{'kak': 0, 'henk': 8}, {'koe': 5, 'henk': 1}]
    >>> find_first(is_henk_1, data)
    {'koe': 5, 'henk': 1}
    >>> find_first(is_henk_1, [])
    Traceback (most recent call last):
      ...
    StopIteration
    """
    return next(filter(function, iterable))


def merge_dicts(
    target, updates, overwrite=False, multivalue=False, deduplicate_iterables=False
):
    """
    Merge two dicts

    >>> first = {'a': 1, 'b': [1], 'c': {'d': 1, 'e': [2]}}
    >>> second = {'a': 3, 'b': 'hanny', 'c': {'d': 9, 'e': [7], 'f': [99]}, 'g': 8}
    >>> merge_dicts(first, second)
    {'a': 1, 'b': [1, 'hanny'], 'c': {'d': 1, 'e': [2, 7], 'f': [99]}, 'g': 8}

    >>> first = {'a': 1, 'b': [1], 'c': {'d': 1, 'e': [2]}}
    >>> merge_dicts(first, second, multivalue=True)
    {'a': [1, 3], 'b': [1, 'hanny'], 'c': {'d': 1, 'e': [2, 7], 'f': [99]}, 'g': 8}

    >>> first = {'a': 1, 'b': [1], 'c': {'d': 1, 'e': [2]}}
    >>> merge_dicts(first, second, overwrite=True)
    {'a': 3, 'b': [1, 'hanny'], 'c': {'d': 9, 'e': [2, 7], 'f': [99]}, 'g': 8}

    >>> first = {'a': 1, 'b': [1]}
    >>> second = {'b': [1]}
    >>> merge_dicts(first, second)
    {'a': 1, 'b': [1, 1]}

    >>> first = {'a': 1, 'b': [1]}
    >>> second = {'b': [1]}
    >>> merge_dicts(first, second, deduplicate_iterables=True)
    {'a': 1, 'b': [1]}

    >>> first = {'a': ["klaas", "piet"], 'b': [3, 1]}
    >>> second = {'a': "piet", 'b': 1}
    >>> merge_dicts(first, second, multivalue=True, deduplicate_iterables=True)
    {'a': ['klaas', 'piet'], 'b': [3, 1]}

    >>> first = {'a': ["klaas", "piet"], 'b': [3, 1]}
    >>> second = {'a': "piet", 'b': 1}
    >>> merge_dicts(first, second, multivalue=True, deduplicate_iterables=False)
    {'a': ['klaas', 'piet', 'piet'], 'b': [3, 1, 1]}
    """
    for key, value in updates.items():
        if key in target:
            item = target[key]
            if isinstance(item, dict):
                if isinstance(value, Mapping):
                    target[key] = merge_dicts(item, value, overwrite)
                elif overwrite:
                    target[key] = value
                else:
                    raise Exception(  # pylint:disable=broad-exception-raised
                        "can not merge nub"
                    )
            elif isinstance(item, list):
                if isinstance(value, str):
                    if deduplicate_iterables and value in item:
                        continue
                    else:
                        item.append(value)
                elif isinstance(value, Iterable):
                    item = item + list(value)
                    if deduplicate_iterables:
                        item = list(unique_everseen(item))
                else:
                    if deduplicate_iterables and value in item:
                        continue
                    else:
                        item.append(value)

                target[key] = item
            elif multivalue and value != item:  # gather multiple values into list
                target[key] = [item, value]
            elif overwrite:
                target[key] = value
        else:
            target[key] = value

    return target



import decimal

__all__ = ["to_money", "round_to_two_dec", "round_to_four_dec"]


def to_money(something):
    """
    Convert something into a decimal with the correct precision

    >>> to_money('3.76')
    Decimal('3.76')
    >>> to_money('1.87765')
    Decimal('1.88')
    >>> to_money(1.8977987987)
    Decimal('1.90')
    """
    return round_to_two_dec(decimal.Decimal(something))


def round_to_two_dec(number):
    """
    Convert a decimal to 2 digit money notation.

    >>> a = decimal.Decimal('3.7555999')
    >>> a
    Decimal('3.7555999')
    >>> round_to_two_dec(a)
    Decimal('3.76')
    >>> round_to_two_dec(a) == decimal.Decimal('3.76')
    True
    """
    return number.quantize(decimal.Decimal("0.01"), decimal.ROUND_HALF_UP)


def round_to_four_dec(number):
    """
    Convert a decimal to 4 digit money notation.

    >>> a = decimal.Decimal('3.7555999')
    >>> a
    Decimal('3.7555999')
    >>> round_to_four_dec(a)
    Decimal('3.7556')
    >>> round_to_four_dec(a) == decimal.Decimal('3.7556')
    True
    """
    return number.quantize(decimal.Decimal("0.0001"), decimal.ROUND_HALF_UP)
